/*==================================================  
       file : supports.c
   
    purpose : auxiliary files
 
  developed : 1991

     author : Kalyanmoy Deb
====================================================*/

#include "mga.ext"
long PrG;

int length_chrom(chr)
struct GENE *chr;
/*  calculates the length of a chromosome  */
{
	int len = 0;
	struct GENE *temp;

	temp = chr;
	/*  NIL chrom has a zero length  */
	if (temp != NIL)
	{
		for (len=1; temp->nextgene != NIL; )
		{
			len++;
			temp = temp->nextgene;
		}
	}
	return(len);
}


ALLELES get_highest_priority_gene(temp)
struct GENE *temp;
{
	GENE_ID Gene;
	ALLELES Allele;
	struct GENE *temp1;
	long Priority;

	Gene = temp->genenumber;
	Allele = temp->allele;
	Priority = temp->priority;
	temp1 = temp->nextgene;
	while (temp1 != NIL) {
if(temp1->genenumber == Gene && temp1->priority>Priority)
   {	Allele = temp1->allele;
	Priority = temp1->priority;
   }
   temp1 = temp1->nextgene;
	}
	if(Priority < 1 )
		Allele = -10;
	free(temp1);
	PrG = Priority;
        return Allele;
}	
					

double fill_chrom(chr)
struct INDIVIDUAL *chr;
/*  fills up the remaining genes from the template  */
{
        register int i;
	BOOLEAN gene_flag[MAX_SIZE];
	struct GENE *temp;
	int id, rem, temp_gene, count = 0;
	unsigned temp_allele, shield, tempmask_gene, tempmask = 1;
	unsigned *rawchr = (*chr).fullchrom;
	unsigned *rawgn = (*chr).fullgene;
	double psum=0.0;

	/*   no gene has been considered yet  */
	for (i = 0; i < problem_length; i++)
		gene_flag[i] = 0;

	/*  assign the template alleles first */
	for (i = 0; i < bytesize; i++)
	{
		rawchr[i] = template[i];
		rawgn[i] = 0;
        }
	temp = (*chr).firstgene;

	/*  for each member in the chromosome  */
	while (temp != NIL) {
		temp_gene = temp->genenumber;   
		temp_allele = temp->allele;

	/*	temp_allele = get_highest_priority_gene(temp);    */

		/*  if not replaced already, (left to right scan)  */
		if (! gene_flag[temp_gene])
		{
			id = temp_gene/UNSINTSIZE;
			rem = temp_gene % UNSINTSIZE;
			shield = tempmask << rem;

			/*  replace the allele, if different  */
/*			if (((rawchr[id] >> rem) & tempmask)  != temp_allele)
				rawchr[id] = shield ^ rawchr[id];   */
/* 2 is the null bit  */ 
if ( ((rawchr[id] >> rem) & tempmask) != temp_allele) {
				rawchr[id] = shield ^ rawchr[id];   
			}
                        /*  make fullgene */
                        tempmask_gene = (0 >> rem) | 1;
			tempmask_gene <<= rem;
			rawgn[id] |= tempmask_gene;
			count++;

			/*  set the allele flag off  */
			gene_flag[temp_gene] = 1;
			psum += (double) PrG;
		}
		temp = temp->nextgene;
	}
	(*chr).genelen = count;
	psum -= count;
	return psum;
}


void copy_chrom(chr1,chr2)
struct INDIVIDUAL chr1,*chr2;
/*  copy chr into another variable whose firstgene has been allocated */
{
	register int i,j;
	struct GENE *temp, *tchr;
	char msg[PAGEWIDTH];

	/*   copy genes   */
	temp = (*chr2).firstgene;
	tchr = chr1.firstgene;
	temp->genenumber = tchr->genenumber;
	temp->allele = tchr->allele;
	temp->priority = tchr->priority;
	for (i = 1; i < chr1.chromlen; i++)
	{
		if (temp->nextgene == NIL)
		{
			if(!(temp->nextgene = (struct GENE *) \
						malloc(sizeof(struct GENE))))
			{
				sprintf(msg,"Insufficient memory for genes\n");
				print_error(msg);
			}
                        temp->nextgene->nextgene = NIL;
		}
		temp = temp->nextgene;
		tchr = tchr->nextgene;
		temp->genenumber = tchr->genenumber;
		temp->allele = tchr->allele;
	temp->priority = tchr->priority;
	}
	/*   copy other members   */
	(*chr2).lastgene = temp;
	temp->nextgene = NIL;
	(*chr2).chromlen = chr1.chromlen;
	for (i = 0; i < bytesize; i++)
	{
		(*chr2).fullchrom[i] = chr1.fullchrom[i];
		(*chr2).fullgene[i] = chr1.fullgene[i];
        }
	(*chr2).genelen = chr1.genelen;
	(*chr2).fitness = chr1.fitness;
/*	(*chr2).Pfitness = chr1.Pfitness;    */
}


BOOLEAN isempty(stack)
STACKPTR stack;
/*  Is the stack empty?  */
{
	return (stack == NIL);
}


BOOLEAN pop_stack(stack, chr)
STACKPTR *stack;
struct INDIVIDUAL *chr;
/*  pops the top member in the stack; sends 1 if there is a member,
    0 otherwise */
{
	STACKPTR t = *stack;

	if (! isempty(t))
	{
		*chr = t->genefirst;
		*stack = t->nextmem;
		free(t);
		return(1);
	}
	else
		return(0);
}


void push_stack(stack, chr)
STACKPTR *stack;
struct INDIVIDUAL chr;
{
        STACKPTR temp;
        char msg[PAGEWIDTH];

        if(!(temp = (STACKPTR)malloc(sizeof(struct STACKTYPE))))
        {
                sprintf(msg,"Insufficient memory for stack\n");
                print_error(msg);
        }
        temp->genefirst = chr;
        temp->nextmem = *stack;
        *stack = temp;
}

void indexx(n,arrin,indx)
int n,indx[];
float arrin[];
{
        int l,j,ir,indxt,i;
        float q;

        for (j=1;j<=n;j++) indx[j]=j;
        l=(n >> 1) + 1;
        ir=n;
        for (;;) {
                if (l > 1)
                        q=arrin[(indxt=indx[--l])];
                else {
                        q=arrin[(indxt=indx[ir])];
                        indx[ir]=indx[1];
                        if (--ir == 1) {
                                indx[1]=indxt;
                                return;
                        }
                }
                i=l;
                j=l << 1;
                while (j <= ir) {
                        if (j < ir && arrin[indx[j]] < arrin[indx[j+1]]) j++;
                        if (q < arrin[indx[j]]) {
                                indx[i]=indx[j];
                                j += (i=j);
                        }
                        else j=ir+1;
                }
                indx[i]=indxt;
        }
}
/*
void push_stack(stack, chr, sf)
STACKPTR *stack;
struct INDIVIDUAL chr;
int sf;
  push a member chr into the stack  
{
	STACKPTR temp;
	char msg[PAGEWIDTH];

	if(!(temp = (STACKPTR)malloc(sizeof(struct STACKTYPE))))
	{
		sprintf(msg,"Insufficient memory for stack\n");
		print_error(msg);
	}
	if(sf == 0) {
	temp->genefirst = chr;
	temp->nextmem = *stack;
	*stack = temp;
	}
	else {
	TRACE("  sf = 1  ");
	fill_chrom(&chr);
	chr.fitness = objfunc(chr.fullchrom);
	dummy_fitness[stack_pos] = chr.fitness;
	shuffle_stack[stack_pos] = stack_pos;
	stack_pos++;
	temp->genefirst = chr;
        temp->nextmem = *stack;
        *stack = temp;
        }

}
*/
void my_sortnum(num,list)
int num;
int *list;
/*  sort an array of integers in ascending order  */
{
        register int i,j;
        int temp;

        for (i = 1; i <= num; i++)
        {
                for (j=i+1; j<=num; j++)
                {
                        if (list[i] > list[j])
                        {
                                temp = list[i];
                                list[i] = list[j];
                                list[j] = temp;
                        }
                }
        }
}
/*
sort_stack()
{
struct INDIVIDUAL chr;
STACKPTR temp;
int i;
	TRACE("  CALLING indexx  ");
	for(i=1; i<stack_pos; i++)
	fprintf(stdout, "i=%d  dummy_fitness=%lf  shuffle_stack = %d\n",i, dummy_fitness[i], shuffle_stack[i]);
	getchar(); 
	getchar();	
	indexx(stack_pos, dummy_fitness, shuffle_stack);	
	for(i=1; i<stack_pos; i++)
	fprintf(stdout, "i=%d  shuffle_stack = %d\n",i, shuffle_stack[i]);
	my_sortnum(stack_pos, shuffle_stack);
}

*/




ALLELES pickallele(id,rawchr)
int id;
unsigned *rawchr;
/*  picks the allele value of the id-th gene from rawchr  */
{
	unsigned mask = 1;
	int num, rem;

	num  = id / UNSINTSIZE;
	rem  = id % UNSINTSIZE;
	if (((rawchr[num] >> rem) & mask) == 1)
		return(1);
	else 
		return(0);
}


void shuffle_pop()
/*  shuffles the shuffle array  */
{
	register int i;
	int num; 
	INDV_ID temp;

	for (i = 0; i < popsize; i++)
		shuffle[i] = i;
	for (i = 0; i < popsize-1; i++)
	{
		num = rnd(i, popsize-1);
		temp = shuffle[num];
		shuffle[num] = shuffle[i];
		shuffle[i] = temp;
	}
}


void assign_beststring_to_template()
/* assigns the best string to the template */
{
	register int i, j;
	long clock,time();
	BOOLEAN gene_flag[MAX_SIZE];
	char *ctime();
	FILE *fp, *fopen();
	struct GENE *temp;
/*	unsigned *rawchr = best_indv.fullchrom;    */
	unsigned *rawchr;   
	unsigned *tmpchr;   
	unsigned temp_allele, tempmask = 1;
	int id, rem, temp_gene;

        rawchr = (unsigned *)malloc(bytesize*sizeof(unsigned));
        tmpchr = (unsigned *)malloc(bytesize*sizeof(unsigned));
	
        for (i = 0; i < problem_length; i++)
                gene_flag[i] = 0;
    /*    for (i = 0; i < bytesize; i++)
		template[i] = best_indv.fullchrom[i];   */
	templatefitness = best_indv.fitness;

/*	writefullchrom(stdout,template);
	fwritechrom(stdout,best_indv,14);
	getchar(); */

	temp = best_indv.firstgene;
	for (j = 0; j < bytesize; j++) {
                rawchr[j] = best_indv.fullchrom[j];  
                tmpchr[j] = template[j];  
	}

        while (temp != NIL) {
                temp_gene = temp->genenumber;
                temp_allele = temp->allele;
		temp_allele = get_highest_priority_gene(temp);
	if (!gene_flag[temp_gene])  {
		id = temp_gene/UNSINTSIZE;
                rem = temp_gene % UNSINTSIZE;

	if (((tmpchr[id] >> rem) & tempmask) != temp_allele) 
		temp_priority[temp_gene] += 1;    
	 /*    if (((rawchr[id] >> rem) & tempmask) != temp_allele)
		temp_priority[temp_gene] += 1; */
		gene_flag[temp_gene] = 1;
	}
		temp = temp->nextgene;
      	} 

        for (i = 0; i < bytesize; i++)
		template[i] = best_indv.fullchrom[i];  
	for(i=0; i<problem_length; i++)
		temp_priority[i] += 0;
 
	/*  write the best string in output file  */
	fp = fopen(Outputfilename,"a");
        fprintf(fp,"\nThe best string:\n");
	writefullchrom(fp,best_indv.fullchrom);
	fprintf(fp,"Fitness = %10.3lf\n",best_indv.fitness);
	fprintf(fp,"String length = %5d\n",best_indv.chromlen);
	fprintf(fp,"Chromosome: ");
	fwritechrom(fp,best_indv,14);
	fprintf(fp,"\n Template priority:\n");
	for(i=0; i<problem_length; i++)
		fprintf(fp, "%d  ", temp_priority[i]);
	fprintf(fp,"\n");

	time(&clock);
	fprintf(fp,"\nEnd of run for era %d: %s\n",era,ctime(&clock));
	if (era != max_era) {
		page(fp);
		fprintf(fp,"\n");
	}
	fclose(fp);
}
